# -*- coding: utf-8 -*-
"""
Editeur de Spyder

Ceci est un script Python du jeu Puissance 4 avec une interface graphique
Pygame. Le joueur 1 a les pions rouges. C'est toujours un joueur humain.
Le joueur 2 a les pions jaunes. C'est un joueur humain ou l'ordinateur.
Le choix entre joueur 2 humain ou joueur 2 ordinateur se fait en modifiant
deux instructions elif.
Voir les indications en commentaires vers la fin du script.

"""

# ----- Importation des modules -----

import sys

import pygame

import premier_module as M01
#  M01 contient les fonctions :
#      affichage_console(matrice)
#      case_libre_la_plus_basse(matrice, num_colonne)

import deuxieme_module as M02
#  M02 contient la fonction :
#  coup_gagnant_horizontal_ou_vertical(matrice, joueur)

import troisieme_module as M03
#  M03 contient la fonction :
#  coup_gagnant_oblique(matrice, joueur)

# Pour desactiver des alertes inapropriees dans Pylint :
# pylint: disable=C0103
# pylint: disable=E1101




#  ----- Constantes utilisees dans ce module -----

#  NB_LIGNE est le nombre de lignes dans la matrice de jeu.
#  NB_COLONNE est le nombre de colonnes dans la matrice de jeu.

#  COTE_CASE est le cote en pixels de chaque case du plateau de jeu.
#  BLEU, NOIR, ROUGE, JAUNE sont des triplets d'intensite de couleurs RVB.
#  RAYON est le rayon des disques noirs, rouges ou jaunes du plateau de jeu.

#  JOUEUR1, JOUEUR2 valent respectivement 1 et 2.
#  JOUEUR3 qui vaut 3 est un joueur fictif pour faire jouer l'ordinateur.

from constantes import NB_LIGNE, NB_COLONNE, COTE_CASE, BLEU, NOIR, RAYON,\
    ROUGE, JAUNE, BLANC, JOUEUR1, JOUEUR2, JOUEUR3



# ----- Declaration des fonctions -----



##############################################################################
#  1. Initialiser une nouvelle partie.
##############################################################################

def nouvelle_partie():
    '''
    Elle renvoie la matrice nulle c'est a dire remplie de 0.
    Elle renvoie aussi le numero du joueur qui commence.
    C'est le 1.


    Returns
    --------
            matrice_vide : TYPE liste de listes d'entiers.
            C'est la matrice de 6 lignes de 7 elements qui est pleine de zeros.

            joueur : TYPE entier.
            C'est le numero du joueur qui commence.

    '''

    matrice_nulle = [[0, 0, 0, 0, 0, 0, 0],\
                     [0, 0, 0, 0, 0, 0, 0],\
                     [0, 0, 0, 0, 0, 0, 0],\
                     [0, 0, 0, 0, 0, 0, 0],\
                     [0, 0, 0, 0, 0, 0, 0],\
                     [0, 0, 0, 0, 0, 0, 0]]
    #
    # Tirage au sort du premier joueur.
    joueur = 1
    #  1 equivaut a JOUEUR1
    #  2 equivaut a JOUEUR2
    return matrice_nulle, joueur



##############################################################################
#  2. Dessiner le plateau de jeu a partir de la matrice qui represente le jeu.
##############################################################################

def dessiner_plateau(matrice):
    """
    Avant l'appel a cette fonction, l'etape 1 de Pygame : definition de l'objet
    Surface "fenetre Pygame" par l'instruction pygame.display.set_mode(taille).

    Cette fonction realise les dessins dans la fenetre (rectangle bleu et ronds
    rouges ou jaunes qui refletent le contenu de la matrice de l'etat du jeu.)


    Parameters
    ----------
    matrice : TYPE liste de listes d'entiers.
        Elle contient des elements 0 (disques noirs) ou 1 (disques rouges) ou
        2 (disques jaunes).

    Returns
    -------
    None.
    Rien n'est renvoye. Elle met seulement a jour l'ffichage dans la
    fenetre Pygame.

    """
    #  Etape 2 de Pygame : Creer les objets images de Surface :
    #  Dessin du rectangle bleu du plateau de jeu.
    pygame.draw.rect(fenetre, BLEU,\
                     (0, COTE_CASE,\
                      NB_COLONNE*COTE_CASE,\
                      (NB_LIGNE+1)*COTE_CASE))
    #   Dessin des 42 ronds noirs.
    for num_col in range (NB_COLONNE):
        for num_ligne in range(NB_LIGNE):
            pygame.draw.circle(fenetre, NOIR,\
           (int(num_col*COTE_CASE+COTE_CASE/2),\
            int(num_ligne*COTE_CASE+COTE_CASE/2+COTE_CASE)),\
            RAYON)

    #   Dessin des ronds rouges si la case vaut 1 dans la matrice (joueur 1).
    #   et jaune si la case vaut 2  (joueur 2).
    for num_col in range (NB_COLONNE):
        for num_ligne in range(NB_LIGNE):
            if matrice[num_ligne][num_col] == 1:
                # L'ordonnee du centre du pion est decalee d'une unite vers le
                # bas a caue de la zone de messages qui occupe le haut de la
                # fenetre. D'ou la formule int((r+1)*LARGEUR_CASE...)
                pygame.draw.circle(fenetre, ROUGE,\
                (int(num_col*COTE_CASE+COTE_CASE/2),\
                int((num_ligne+1)*COTE_CASE+COTE_CASE/2)),\
                RAYON)
            if matrice[num_ligne][num_col] == 2:
                pygame.draw.circle(fenetre, JAUNE,\
                (int(num_col*COTE_CASE+COTE_CASE/2),\
                int((num_ligne+1)*COTE_CASE+COTE_CASE/2)),\
                RAYON)
    #
    #  Etape 3 de Pygame :	Preparer le dessin des objets images de Surface
    #  crees a l'etape 2 sur l'objet de Surface d'affichage cree a l'etape 1.
    #  L'instructione est
    #  fenetre.blit(<nom de l'objet image de Surface>, (coordonnees))
    #  REMARQUE : ici, a l'etape 2 on a utilise Le module draw qui est le
    #  module Pygame qui permet d'afficher des formes geometriques.
    #  On n'a pas utilise a l'etape 2 le module image qui sert a afficher des
    #  fichiers de type image. Donc l'etape 3 est inutile.
    #
    #
    #  Etape 4 de Pygame : Affichage de la fenetre (on peut aussi utiliser
    #  pygame.display.update() si on veut raffraichir la fenetre entiere
    #  en une seule fois).
    pygame.display.flip()







##############################################################################
# 3. Verifie que la case choisie est jouable, c'est a dire qu'elle contient
#    un zero.
##############################################################################

def case_jouable(matrice, ligne, colonne):
    """
    Cette fonction teste si la case ayant la ligne et la colonne passees en
    parametres est libre. Une case est libre lorsqu'elle contient 0.


    Parameters
    ----------
    matrice : TYPE liste de listes d'entiers.
        Elle memorise l'etat du jeu a chaque instant.

    colonne : TYPE entier.
        C'est le numero de la colonne dont on veut savoir si elle contient
        encore au moins une case vide.

    Returns
    -------
    resultat : TYPE booleen.
        S'il n'y a plus de place dans la colonne passee en parametre alors le
        resultat renvoye est False.

        S'il a encore de la place dans la colonne passee en parametre alors le
        resultat renvoye est True.

    """
    resultat = False
    if matrice[ligne][colonne] == 0:
        resultat = True

    return resultat


##############################################################################
#  4. Pour placer le numero du joueur dans la colonne choisie,
#     sur la ligne libre.
##############################################################################

def jouer_coup(matrice, joueur, ligne, colonne):
    """
    Cette fonction renvoie False si la case joue par le joueur n'est pas
    jouable c'est a dire qu'elle n'est pas une case vide.
    Dans le programme principal, cela empechera de passer le tour
    a l'autre joueur. Dans ce cas ce sera toujours au meme joueur de jouer.

    S'il existe une case libre dans la colonne choisie par le joueur alors
    cette fonction renvoie True et alors ce sera a l'autre joueur de jouer.

    Parameters
    ----------
    matrice : TYPE liste de listes d'entiers.
        Elle memorise l'etat du jeu a chaque instant.
    joueur : TYPE entier.
        C'est le numero du joueur qui est en train de jouer. Il peut etre 1 ou 2.
    ligne : TYPE entier.
        C'est le numero de ligne joue par le joueur qui est en train de jouer.
    colonne : TYPE entier.
        C'est le numero de colonne joue par le joueur qui est en train de jouer.

    Returns
    -------
    nouveau_coup_inscrit : TYPE booleen.
        Vaut True si le coup est valide, False autrement. Il est False si
        le joueur clique sur une colonne qui n'a pas de case libre.


    """
    nouveau_coup_inscrit = False

    #  La case est-elle jouable ?
    if case_jouable(matrice, ligne, colonne):
        #  case_libre est le tuple (numero de ligne, numero de colonne) de
        #  la case vide la plus basse dans la colonne indiquee.
        #  Si la fonction case_libre_la_plus_basse contient "pass" alors
        #  elle renvoie "None".
        case_libre = M01.case_libre_la_plus_basse(matrice, colonne)
        if case_libre is not None :
            ligne, colonne = M01.case_libre_la_plus_basse(matrice, colonne)

        matrice[ligne][colonne] = joueur
        nouveau_coup_inscrit = True

    #  Si aucune case n'est libre (c'est a dire la colonne n'est pas jouable)
    #  alors la fonction renvoie False.
    return nouveau_coup_inscrit



##############################################################################
#  5. Pour detecter si le joueur passe en parametre a un alignement gagnant
#  dans la matrice passee en parametre.
##############################################################################

def coup_gagnant(matrice, joueur):
    '''
    Cette fonction renvoie True si 4 pions de meme couleur sont alignes
    1) verticalement ou horizontalement
        ou
    2) en oblique.
    Elle renvoie False sinon.

    Parameters
    ----------
        matrice : TYPE liste de listes d'entiers.
                 Elle represente le plateau du jeu de 6 lignes et 7 colonnes.

        joueur : TYPE entier.
                 Vaut 1 pour verifier si le joueur 1 a un alignement gagnant.
                 Vaut 2 pour verifier si le joueur 2 a un alignement gagnant.


    Returns
    -------
        resultat : TYPE booleen.
                   resultat vaut True si 4 pions de meme couleur sont alignes
                   verticalement ou horizontalement ou en oblique.
                   resultat vaut False dans le cas contraire.

    '''
    resultat = M02.coup_gagnant_horizontal_ou_vertical(matrice, joueur) or\
    M03.coup_gagnant_oblique(matrice, joueur)
    return resultat



##############################################################################
#  6. Pour tester s'il reste encore au moins une case vide et ainsi permettre
#  de detecter une partie nulle eventuelle.
##############################################################################

def partie_nulle(matrice):
    """
    Cette fonction renvoie True si a matrice est pleine. Vu la structure du
    programme principal, cette fonction n'est appelee qu'apres avoir verifie
    qu'il n'y a pas d'alignement gagnant. Cela signifie que la detection
    d'une matrice pleine equivaut a la detection d'une partie nulle.
    Elle renvoie False sinon.

    Parameters
    ----------
        matrice : TYPE liste de listes d'entiers.
                 Elle represente le plateau du jeu de 6 lignes et 7 colonnes.

    Returns
    -------
        resultat : TYPE booleen.
                   resultat vaut True si la matrice est pleine et donc que la
                   partie est nulle.
                   resultat vaut False dans le cas contraire.

    """
    resultat = True
    for num_ligne in range (NB_LIGNE):
        for num_col in range(NB_COLONNE):
            if matrice[num_ligne][num_col] == 0:
                resultat = False
    return resultat




##############################################################################
#  Fonction ordi_joue(matrice) qui genere le choix de colonne
#  fait par l'ordinateur lorsqu'on veut jouer contre la machine.
#  La ligne choisie par l'ordinateur est toujours la numero 0. Celle-ci est
#  corrigee par le fonction M01.case_libre_la_plus_basse(matrice, num_colonne)
#  qui doit etre operationnelle.
##############################################################################

##############################################################################
# 7a.Fonction necessaire a la fonction ordi_joue(matrice).
#    Elle verifie que la colonne choisie est jouable, c'est a dire qu'elle
#    contient un zero en haut de la colonne.
##############################################################################

def colonne_jouable(matrice, colonne):
    """
    Cette fonction teste si la case en haut de la colonne contient 0 c'est
    a dire que la colonne a encore de la place.


    Parameters
    ----------
    matrice : TYPE liste de listes d'entiers.
        Elle memorise l'etat du jeu a chaque instant.
    colonne : TYPE entier.
        C'est le numero de la colonne dont on veut savoir si elle contient
        encore au moins une case vide.

    Returns
    -------
    resultat : TYPE booleen.
        S'il n'y a plus de place dans la colonne passee en parametre alors le
        resultat renvoye est False.

        S'il a encore de la place dans la colonne passee en parametre alors le
        resultat renvoye est True.

    """
    resultat = False
    if matrice[0][colonne] == 0:
        resultat = True

    return resultat


##############################################################################
# 7b.Trouver une case de facon strategique pour le JOUEUR2 lorsque c'est un
#   joueur artificiel, c'est a dire l'ordinateur.
##############################################################################

def ordi_joue(matrice):
    """
    Une copie de la matice est faite afin de tester les choix sans modifier
    la matrice de jeu dans son etat actuel.
    Elle sera modifiee a l'exterieur de la fonction.
    L'ordinateur choisit un numero de colonne en fonction des trois priorites
    suivantes classees de la plus prioritaire a la moins prioritaire :
        1) Le numero de colonne correspond a un coup gagnant pour l'ordinateur
    (qui a toujours le role de joueur 2).
        2) Le numero de colonne correspond a un coup gagnant pour le joueur
    adverse (qui a toujours le role de joueur 1).
        3) le numero de colonne est un choix aleatoire parmi les numeros de
    colonne encore jouables.

    Parameters
    ----------
        matrice : TYPE liste de listes d'entiers.
                 Elle represente le plateau du jeu de 6 lignes et 7 colonnes.

    Returns
    -------
    num_col : TYPE entier
        C'est le numero de colonne joue par l'ordinateur.

    """
    pass





# -----------------------------Partie Principale -----------------------------

LARGEUR_FENETRE = NB_COLONNE * COTE_CASE
HAUTEUR = (NB_LIGNE+1) * COTE_CASE
TAILLE = (LARGEUR_FENETRE, HAUTEUR)

# Initialisation de Pygame
pygame.init()


#  Etape 1 de Pygame : Creation d'un objet de Surface d'affichage (fenetre).
#  Cette etape a lieu en dehors de la fonction dessiner_plateau(matrice)
#  ou se trouvent les etapes de construction des dessins Pygame et de
#  la mise a jour de la fenetre.
#
#  l'objet fenetre est ainsi disponible dans la boucle infinie de jeu.
fenetre = pygame.display.set_mode(TAILLE)

#  Ecriture du titre dans la fenetre.
pygame.display.set_caption("Puissance 4")

#  Initialisation de matrice par une matrice nulle et de joueur par 1
matrice_du_jeu, joueur_actif =  nouvelle_partie()

#  Double affichage : Matrice dans la console + plateau de jeu Pygame.
M01.affichage_console(matrice_du_jeu)
dessiner_plateau(matrice_du_jeu)


#  ----------------Boucle infinie de jeu ------------------------
JEU_TERMINE = False

while not JEU_TERMINE:
    for event in pygame.event.get():
        if event.type==pygame.QUIT:
            JEU_TERMINE = True
            pygame.quit()
            sys.exit()

        #  Cas ou le joueur actif est le JOUEUR1.
        if joueur_actif == JOUEUR1:
            #  Effacer la premiere ligne du plateau.
            pygame.draw.rect(fenetre,NOIR,(0,0,LARGEUR_FENETRE,COTE_CASE))
            myfont = pygame.font.SysFont("comicsansms", 40)
            texte_image = myfont.render("Le joueur 1 joue", True, ROUGE)
            fenetre.blit(texte_image, [100,10])
            pygame.display.update()

            if event.type == pygame.MOUSEBUTTONDOWN:
                #  Capture de l'abscisse de la souris au moment du clic.
                posx = event.pos[0]
                posy = event.pos[1]
                #  Conversion de l'ordonne de la souris en numero de ligne.
                ligne_choisie = posy//COTE_CASE
                #  Correction de la ligne choisie pour tenir compte que la
                #  zone de messages en haut de la fenetre occupe les ordonnees
                #  allant de 0 a 100.
                ligne_choisie = ligne_choisie - 1
                #  Conversion de l'abscisse de la souris en numero de colonne.
                colonne_choisie = posx//COTE_CASE
                choix = ligne_choisie, colonne_choisie
                #  Enregistre le coup du joueur s'il est dans une colonne valide.
                nouveau_coup_execute = jouer_coup(matrice_du_jeu,\
                                                  JOUEUR1,\
                                                  choix[0],\
                                                  choix[1])
                M01.affichage_console(matrice_du_jeu)
                dessiner_plateau(matrice_du_jeu)
                if coup_gagnant(matrice_du_jeu, JOUEUR1):
                    #  Effacer la premiere ligne du plateau.
                    pygame.draw.rect(fenetre,NOIR,(0,\
                                                   0,\
                                                   LARGEUR_FENETRE,\
                                                   COTE_CASE))
                    myfont = pygame.font.SysFont("comicsansms", 40)
                    texte_image = myfont.render("Le joueur 1 gagne !",\
                                                True,\
                                                ROUGE)
                    fenetre.blit(texte_image, [100,10])
                    pygame.display.update()
                    pygame.time.wait(3000)
                    JEU_TERMINE = True
                else:
                    if partie_nulle(matrice_du_jeu):
                        #  Effacer la premiere ligne du plateau.
                        pygame.draw.rect(fenetre,NOIR,(0,\
                                                       0,\
                                                       LARGEUR_FENETRE,\
                                                       COTE_CASE))
                        myfont = pygame.font.SysFont("comicsansms", 40)
                        texte_image = myfont.render("Partie nulle",\
                                                    True,\
                                                    BLANC)
                        fenetre.blit(texte_image, [100,10])
                        pygame.display.update()
                        pygame.time.wait(3000)
                        JEU_TERMINE = True
                    else:
                        pygame.time.wait(1000)
                        #  En cas de colonne choisie valide,
                        #  alors le jeu continue avec changement de joueur.
                        if nouveau_coup_execute is True :
                            joueur_actif = JOUEUR2
                        else:
                            # Sinon on reste avec le meme joueur.
                            joueur_actif = JOUEUR1


        #  Cas ou le joueur 2 est humain : saisir ci-dessous
        #  joueur_actif == JOUEUR2:
        #  Cas ou le joueur 2 est l'ordi : saisir ci-dessous
        #  joueur_actif == JOUEUR3:
        elif joueur_actif == JOUEUR2:
            #  Effacer la premiere ligne du plateau.
            pygame.draw.rect(fenetre,NOIR,(0,0,LARGEUR_FENETRE,COTE_CASE))
            myfont = pygame.font.SysFont("comicsansms", 40)
            texte_image = myfont.render("Le joueur 2 joue", True, JAUNE)
            fenetre.blit(texte_image, [100,10])
            pygame.display.update()

            if event.type == pygame.MOUSEBUTTONDOWN:
                #  Capture de l'abscisse de la souris au moment du clic.
                posx = event.pos[0]
                posy = event.pos[1]
                #  Conversion de l'ordonne de la souris en numero de ligne.
                ligne_choisie = posy//COTE_CASE
                #  Correction de la ligne choisie pour tenir compte que la
                #  zone de messages en haut de la fenetre occupe les ordonnees
                #  allant de 0 a 100.
                ligne_choisie = ligne_choisie - 1
                #  Conversion de l'abscisse de la souris en numero de colonne.
                colonne_choisie = posx//COTE_CASE
                choix = ligne_choisie, colonne_choisie
                #  Enregistre le coup du joueur s'il est dans une colonne valide
                nouveau_coup_execute = jouer_coup(matrice_du_jeu,\
                                                  JOUEUR2,\
                                                  choix[0],\
                                                  choix[1])
                M01.affichage_console(matrice_du_jeu)
                dessiner_plateau(matrice_du_jeu)
                if coup_gagnant(matrice_du_jeu, JOUEUR2):
                    #  Effacer la premiere ligne du plateau.
                    pygame.draw.rect(fenetre,NOIR,(0,\
                                                   0,\
                                                   LARGEUR_FENETRE,\
                                                   COTE_CASE))
                    myfont = pygame.font.SysFont("comicsansms", 40)
                    texte_image = myfont.render("Le joueur 2 gagne !",\
                                                True,\
                                                JAUNE)
                    fenetre.blit(texte_image, [100,10])
                    pygame.display.update()
                    pygame.time.wait(3000)
                    JEU_TERMINE = True
                else:
                    if partie_nulle(matrice_du_jeu):
                        #  Effacer la premiere ligne du plateau.
                        pygame.draw.rect(fenetre,NOIR,(0,\
                                                       0,\
                                                       LARGEUR_FENETRE,\
                                                       COTE_CASE))
                        myfont = pygame.font.SysFont("comicsansms", 40)
                        texte_image = myfont.render("Partie nulle",\
                                                    True,\
                                                    BLANC)
                        fenetre.blit(texte_image, [100,10])
                        pygame.display.update()
                        pygame.time.wait(3000)
                        JEU_TERMINE = True
                    else:
                        pygame.time.wait(1000)
                        #  En cas de colonne choisie valide,
                        #  alors le jeu continue avec changement de joueur.
                        if nouveau_coup_execute is True :
                            joueur_actif = JOUEUR1
                        else:
                            # Sinon on reste avec le meme joueur.
                            joueur_actif= JOUEUR2

        #  Cas ou le joueur 2 est humain : saisir ci-dessous
        #  joueur_actif == JOUEUR3:
        #  Cas ou le joueur 2 est l'ordi : saisir ci-dessous
        #  joueur_actif == JOUEUR2:
        elif joueur_actif == JOUEUR3:
            #  Effacer la premiere ligne du plateau.
            pygame.draw.rect(fenetre,NOIR,(0,0,LARGEUR_FENETRE,COTE_CASE))
            myfont = pygame.font.SysFont("comicsansms", 40)
            texte_image = myfont.render("L'ordinateur joue", True, JAUNE)
            fenetre.blit(texte_image, [100,10])
            pygame.display.update()

            if not JEU_TERMINE:
                choix = ordi_joue(matrice_du_jeu)
                #  Enregistre le choix de colonne fait par l'ordinateur.
                #  L'ordinateur joue toujours dans la ligne 0.
                #  Si la fonction ordi_joue(matrice_du_jeu) est operationnelle.
                if choix is not None:
                    nouveau_coup_execute = jouer_coup(matrice_du_jeu,\
                                                      JOUEUR2,\
                                                      0,\
                                                      choix)
                else:
                    joueur_actif = JOUEUR1

                M01.affichage_console(matrice_du_jeu)
                dessiner_plateau(matrice_du_jeu)
                if coup_gagnant(matrice_du_jeu, JOUEUR2):
                    #  Effacer la premiere ligne du plateau.
                    pygame.draw.rect(fenetre,NOIR,(0,\
                                                   0,\
                                                   LARGEUR_FENETRE,\
                                                   COTE_CASE))
                    myfont = pygame.font.SysFont("comicsansms", 40)
                    texte_image = myfont.render("L'ordinateur gagne !",\
                                                True,\
                                                JAUNE)
                    fenetre.blit(texte_image, [100,10])
                    pygame.display.update()
                    pygame.time.wait(3000)
                    JEU_TERMINE = True
                else:
                    if partie_nulle(matrice_du_jeu):
                        #  Effacer la premiere ligne du plateau.
                        pygame.draw.rect(fenetre,NOIR,(0,\
                                                       0,\
                                                       LARGEUR_FENETRE,\
                                                       COTE_CASE))
                        myfont = pygame.font.SysFont("comicsansms", 40)
                        texte_image = myfont.render("Partie nulle",\
                                                    True,\
                                                    BLANC)
                        fenetre.blit(texte_image, [100,10])
                        pygame.display.update()
                        pygame.time.wait(3000)
                        JEU_TERMINE = True
                    else:
                        pygame.time.wait(1000)
                        #  En cas de colonne choisie valide,
                        #  alors le jeu continue avec changement de joueur.
                        if nouveau_coup_execute is True :
                            joueur_actif = JOUEUR1
                        else:
                            # Sinon on reste avec le meme joueur.
                            joueur_actif = JOUEUR2


#  Partie qui gere l'apres sortie de boucle infinie de jeu.
pygame.quit()
sys.exit()
